/* this is the LangstonMerge in-place merging algorithm sent to me by
   Prof. John Ellis (jellis@csr.csc.UVic.CA). 

   Note that this implementation uses a O(N) bookkeeping array moved[]
   and thus isn't really in-place */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

static int *A;
static char *moved;
static int move, comp;



static int partition( int i, int j ) {
    int test, temp;

    i--;
    j++;
    test = A[(j+i) >> 1];
    for( ; ; ) {
        while(A[++i] < test) ;
           /*  comp ++; */
        while(A[--j] > test) ;
           /*  comp ++; */
        /* comp += 2;  */
        if (j <= i)
            break;
        temp = A[i];
        A[i] = A[j];
        A[j] = temp;
        /* move += 3; */
    }
    return j;
}


static void Qsort( int i, int j ) {
    int k;

    if (i < j) {
       k = partition(i, j);
       Qsort(i, k);
       Qsort(k+1, j);
    }
}


/************************************************************************/
/* Rotate : Cyclic shift the array elements between and including       */
/* A[start] and A[end] by num places to the right                       */
/* Input: start: array start                                            */
/*        end: array end                                                */
/*        num : amount of shift right                                   */
/************************************************************************/
static int Rotate(int start,int end, int num)

{
int p, n, c, q, temp, r;

if(start==end) return 0;


p = start - 1;  n = end - start+1;  c = 0;

while (c < n)
  { p = p+1;  q = p; temp = A[q];
    for ( ; ; )
	{ if (q > start + num - 1) r = q -  num;
	     else r = q + n - num;
    if (r == p) break;
if((r<0) ||(p<0)) break;
	A[q] = A[r]; q = r; c++;
	}
    A[q] = temp; c++;
  }
return 0;
}


/************************************************************************/
/* SubMerge : merge two sets together using a buffer                    */
/* Input: start1: start of first set                                    */
/*          end1:  end of first set                                     */
/*        start2: start of second set                                   */
/*          end2:  end of second set                                    */
/*       buffindex: start index of the buffer                           */
/* Modify:                                                              */
/*        newstart: new start for next merge                            */
/*        bufferindex: new buffer index location                        */
/************************************************************************/

static void SubMerge(int start1, int end1, int start2, int end2,int * buffindex)
{
 /* must be returned back */
int index1,index2,index3;
int temp;
index1 = start1;
index2 = start2;
index3 = *buffindex;

while(index1<=end1)
	{
  comp++;
  if(A[index1]<=A[index2])
		{
		temp = A[index1];
		A[index1] = A[index3];
		A[index3]= temp;
		index1++;
		index3++;
		move+=3;
		}
	else
		{
		temp = A[index2];
		A[index2] = A[index3];
		A[index3]= temp;
		index2++;
		index3++;
		move+=3;
		}
	}
*buffindex = index3;
}


/************************************************************************/
/* SwapSets : swap two identical-size blocks                            */
/* Input: index1: first block start                                     */
/*        index2: second block start                                    */
/*        size  : common block size                                     */
/************************************************************************/

static void SwapSets(int index1,int index2,int size)
{
int i;
int temp;
for(i=0;i<size;i++)
	{
	temp = A[index1];
	A[index1] = A[index2];
	A[index2] = temp;
	index1++;
	index2++;
	move+=3;
	}
}



/************************************************************************/
/* putmax : put the maxium-value item(s) at the end of the array        */
/* Input: end1  end of the first set                                    */
/*        end1 end of the second set                                    */
/*        reminder: number of items to be moved                         */
/************************************************************************/
static void putmax(int start,int end1,int end2,int reminder)
{
int index1,index2;
int temp;
int i;
int temp2;
index1 = end1;
index2 = end2;
while((reminder>0)&&(index2>end1)&&(index1>=start))
	{
if(A[index1]>A[index2])
		{
		temp =A[index2];
		A[index2] = A[index1];
		i = index1-1;
	while ((A[i]>temp)&&(i>=start))
		{
A[i+1] = A[i];
i--;
		}
		A[i+1] = temp;
		index2--;
		/* end1++; */
		}
	else
		{
		index2--;
		}
reminder--;
	}

if(reminder>0)
	{
while((index1>=0)&&(reminder>0))
		{
index1--;
reminder--;
		}

while((index2>index1)&&(reminder>0))
		{
index2--;
reminder--;
		}
	}

/* if(reminder>0) printf("\n reminder error\n"); */
}



/************************************************************************/
/* MergeItem : merge an item into a set                                 */
/* Input: start1 start index of the set                                 */
/*        end1 end index of the set                                     */
/*        i index of the item to be merged                              */
/************************************************************************/
static int MergeItem(int i,int start1,int end1)
{
int j;
int temp;
j=end1;
if(A[i]>=A[j]) return -1;
while((A[i]<A[j]) &&(j>=start1)) j--;
j++;

while(j<=end1)
	{
	temp = A[i];
	A[i] = A[j];
	A[j] = temp;
	j++;
	}
return 0; 
}

/************************************************************************/
/* Junction_Sort : sort  the junction block                             */
/* Input: start start index of the first set                            */
/*        end1 end index of the first set                               */
/*        end2 end index of the junction block                          */
/************************************************************************/

static void Junction_Sort(int start1,int end1,int end2)
{
int start2;
int i,flag;
start2 = end1+1;
/* if(out !=NULL) fprintf(out,"\n before junction sorting, start1=%d,
                  end1=%d, start2=%d, end2=%d",start1,end1,start2,end2);
*/
for(i=start2;i<=end2;i++)
	{
	flag = MergeItem(i,start1,end1);
	if(flag==-1) break;
end1++;
}
}


/************************************************************************/
/* Lmerge : merge two sets, using the Langston method                   */
/* Input: start start index of the first set                            */
/*        end1 end index of the first set                               */
/*        end2 end index of the second set                              */
/************************************************************************/
static void Lmerge(int start,int end1,int end2)
{
int s;
int n;
int index,index1,index2,num;
int setnum,reminder;
int i,j,tail1,tail2;
int buffstart;

/*  printf("\n merge from start1 = %d, end1 = %d, end2 = %d",start,end1,end2); */
n = end2-start+1;
s = (int) floor(sqrt((double)n));
setnum = (n)/s;
reminder =(n)%s;

putmax(start,end1,end2,reminder);
num = s;
end2 -=reminder;
if(end2<=end1) { /* printf("\n list"); */return; }

index1 = end1;
index2 = end2;

while((num>0)&&(index1>=0))
	{
/*printf("\n index1=%d, value = %d, index2=%d, value = %d",
          index1,A[index1],index2,A[index2]); */
	if(A[index1]>A[index2])
		index1--;
	else
		index2--;
	num--;

	}
if(index2<=index1) { printf("\n one first list "); return ;}
while(num>0)
	{
num--;
/*index2--; */
	}
Rotate(end1+1,end2,end2-index2);
Rotate(start,end1+end2-index2,s);
i=0;
end1 = end1+end2-index2;
if(end1==end2)
{
Qsort(start,start+s-1); /* sort buffer */
SwapSets(start,s+start,s);
return ;
}
/* if(out!=NULL)  fprintf(out,"\n buffer extraction , block size (s) = %d",s); */

i=(end1-start)/s;
i++;
i*=s;
i+=start;

if(s>1) Junction_Sort(s+start,end1,i-1);
/* set sorting */
for(i=1;i<setnum;i++)
	{
index1 = i*s+start;
tail1=index1+s-1;
index = index1;
for(j=i+1;j<setnum;j++)
	{
index2 = j*s+start;
tail2 = index2+s-1;
comp++;
if(A[tail2]<A[tail1]) { index = index2; tail1=index+s-1; }
	}
if(index != index1) SwapSets(index1,index,s);
	}

index1 = s+start;
index2 = index1+s;
tail1 = index1+s-1;
tail2 = index2+s-1;
buffstart = start;

do
{
 comp+=1;
 while((A[tail1]<=A[index2])&&(tail2<=end2))
  {
   comp+=1;
   index2+=s;
   tail1+=s;
   tail2+=s;
  }
  if(tail1>=end2) break;
  SubMerge(index1, tail1, index2, tail2,&buffstart);
  index1=buffstart+s;
  tail1 = tail2;
  index2=tail1+1;
  tail2 = tail1+s;
} while(tail2<=end2);


if(tail1 > end2 ) tail1 = end2-1;
Rotate(buffstart,end2,end2-index1+1);
buffstart +=tail1-index1+1;
Qsort(buffstart,buffstart+s-1);
}

/* merge two sorted lists of elements into one list. The two lists are
   assumed to be contiguous with the first starting at data[0] and the
   second starting at data[n1]. The two lists are of length n1 and n2
   respectively */
int LangstonMerge(int *data, int n1, int n2)
{
	comp = 0;
	A = data;
	
	/* need some bookkeeping room. */
	moved = (char *)malloc(n1+n2);
	memset(moved, 0, n1+n2);

	Lmerge(0, n1-1, (n1+n2)-1);

	free(moved);
	return comp;
}
