/* this is the ShuffleMerge in-place merging algorithm sent to me by
   Prof. John Ellis (jellis@csr.csc.UVic.CA). 

   Note that this implementation uses a O(N) bookkeeping array moved[]
   and thus isn't really in-place */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static int *A;
static char *moved;
static int move, comp;

#define COMPARE(a,b) (comp++, (a)<(b))

static void pack( int s, int e ) {
    int n, i, temp, p, q, r, count;

    n = e - s + 1;
    count = 0;
    for( i = s;  i <= e+1;  i++ )
       moved[i] = 0;
    p = s-1;

    for( ; ; ) {
	while(moved[++p])
	    ;
	if (p > e)
	    break;
	q = p;
	temp = A[q];
	count++;

	for( ; ; ) {
            if ((q-s+1) % 2 == 1)
                 r =  n/2 + (q-s)/2 + s;  
            else r =  (q-s+1)/2 + s -1;
	    moved[q] = 1;
	    if (r == p)
		break;
	    A[q] = A[r];
	    q = r;
	}

	A[q] = temp;
    }
    move += (e - s + 1) + count;
}

static void unpack( int s, int e ) {
    int n, i, temp, p, q, r, count;

    n = e - s + 1;
    count = 0;
    for( i = s;  i <= e+1;  i++ )
       moved[i] = 0;
    p = s-1;

    for( ; ; ) {
        while(moved[++p])
            ;
        if (p > e)
            break;
        q = p;
        temp = A[q];
        count++;

        for( ; ; ) {
            r = 2 * (q-s+1) % (n + 1) + s-1;
            moved[q] = 1;
            if (r == p)
                break;
            A[q] = A[r];
            q = r;
        }

        A[q] = temp;
    }
    move += (e - s + 1) + count;
}



static void Smerge(int s1, int e1, int e2 ) {
    int p, q, r, e;

/*    if (e1 - s1 >= 0 && e2 - e1 >= 1) */ {
       if (e1-s1 < e2-e1-1) {
          e = 2*e1 - s1 + 1;
          q = e;
          r = q-1;
          p = e2+1;
          pack(s1+1, e-1);
          while(q > s1) {
             while(r >= s1 && COMPARE(A[q],A[r])) {
                r -= 2;
             }
             if (r < q-1) {
                unpack(r+2, q);
                Smerge((q+r+3) / 2, q, p-1);
                p = (q+r+1) / 2;
                q = r+1;
             } else {
                q--;
                r = q-1;
                p = q;
             }
          }
       } else {
          e = 2*e1 - e2 + 1;
          q = e;
          r = q+1;
          p = s1-1;
          pack(e+1, e2-1);
          while(q < e2) {
             while(r <= e2 && COMPARE(A[r],A[q])) {
                r += 2;
             }
             if (r > q+1) {
                unpack(q, r-2);
                Smerge(p+1, q-1, (r+q-3) / 2);
                p = (r+q-1) / 2;
                q = r-1;
             } else {
                q++;
                r = q+1;
                p = q;
             }
          }
       }
    }
}


/* merge two sorted lists of elements into one list. The two lists are
   assumed to be contiguous with the first starting at data[0] and the
   second starting at data[n1]. The two lists are of length n1 and n2
   respectively */
int ShuffleMerge(int *data, int n1, int n2)
{
	comp = 0;
	A = data;
	
	/* need some bookkeeping room. */
	moved = (char *)malloc(n1+n2);
	memset(moved, 0, n1+n2);

	Smerge(0, n1-1, (n1+n2)-1);

	free(moved);
	return comp;
}

