#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <sys/resource.h>


/* call this when out of memory */
static void oom(void)
{
	printf("out of memory\n");
	exit(1);
}

/* create a random list of integers of length n */
static int *rand_list(int n)
{
	int *ret = (int *)malloc(n * sizeof(int));
	int i;

	if (!ret) oom();
	for (i=0;i<n;i++)
		ret[i] = random();
	return ret;
}


/* used by qsort */
static int int_comparison(int *i1, int *i2)
{
	return *i1 - *i2;
}

/* a simple sort routine using qsort - used to establish the lists */
static void sort_list(int *data, int n)
{
	qsort(data, n, sizeof(int), int_comparison);
}


static void usage(void)
{
	printf("testmerge <start_n> <end_n> <multiplier>\n");
}


int ShuffleMerge2(int *data, int n1, int n2);
int dmerge(int *data, int n1, int n2);

FILE *dbf;

int main(int argc, char *argv[])
{
	int *data, *data2;
	int high_n, n, loops=0, e=0, c=0;
	dbf = stderr;

	if (argc < 2) {
		usage();
		exit(0);
	}

	high_n = atoi(argv[1]);

	while (1) {
		n = (((unsigned)random()) % high_n);
		data = rand_list(n);
		sort_list(data, n/2);
		sort_list(data+n/2, n-n/2);

		data2 = malloc(sizeof(int)*n);
		if (!data2) oom();
		
		memcpy(data2, data, sizeof(int)*n);
		dmerge(data2, n/2, n-n/2);
		c += ShuffleMerge2(data, n/2, n-n/2);
		e += n;

		/* make sure they all got the same result! */
		if (memcmp(data, data2, sizeof(int)*n) != 0) {
			printf("data and data2 don't match! (loops=%d n=%d)\n", 
			       loops, n);
			exit(1);
		}

		free(data);
		free(data2);

		loops++;
		if (loops % 100 == 0) {
			printf("%d %g\n", loops, c/(1.0*e));
		}
	}
	
	return 0;
}
